<?php
/*
Plugin Name: Captcha Free Anti Spam for Contact Form 7 (Simple No-Bot)
Plugin URI: http://www.lilaeamedia.com/simple-no-bot/
Description: Simple, lightweight, no captcha, no configuration. Just works.
Version: 2.1.5
Author: Lilaea Media
License: GPLv2
(c) 2019 J Fleming Lilaea Media LLC
*/

if ( !class_exists( 'SimpleNoBot' ) ):

class SimpleNoBot {
    
    private $eventhash;
    private $token;
    
    /**
     * define WP action/filter hooks
     */
    function __construct(){
        //add_action( 'init', array( $this, '_set_cookie' ) );
        add_action( 'wp_enqueue_scripts', array( $this, 'enqueue' ) );
        add_action( 'wp_ajax_nopriv_snb_get_token', array( $this, 'get_token' ) );
        add_action( 'wp_ajax_snb_get_token', array( $this, 'get_token' ) );
        add_filter( 'wpcf7_spam', array( $this, 'validate_session' ) );
        // v.2.0 filter for any form
        add_filter( 'snb_test_spam', array( $this, 'validate_session' ), 10, 2 );
        add_action( 'w3tc_flush_all', array( $this, 'reset_spam_ips' ) );
        if ( is_admin() && isset( $_REQUEST[ 'snb_flush' ] ) )
            add_action( 'init', array( $this, 'reset_spam_ips' ) );
    }
    
    /**
     * set cookie for additional analysis (not currently used)
     */
    function _set_cookie( $flush = FALSE ) {
        if ( wp_doing_ajax() ):
            // dont update cookie for ajax calls, only full page loads
            //$this->log_debug( __METHOD__, 'doing ajax' );
        else:
            //$this->log_debug( __METHOD__, 'setting cookie' );
            if ( empty( $_COOKIE[ 'snb' ] ) ):
                $start = time();
                $loads = 1;
            else:
                $cookieparts = explode( ':', $_COOKIE[ 'snb' ] );
                $start = $cookieparts[ 0 ];
                $loads = empty( $cookieparts[ 1 ] ) ? 1 : $cookieparts[ 1 ] + 1;
            endif;
            setcookie(
                'snb', 
                $start . ':' . $loads, 
                time() + 3600, 
                COOKIEPATH, 
                COOKIE_DOMAIN, 
                FALSE
            );
        endif;
    }

    /**
     * load the script and set locale vars
     */
    function enqueue(){
        wp_enqueue_script( 'snbvars', plugin_dir_url( __FILE__ ) . 'snb' . ( defined( 'SCRIPT_DEBUG' ) && SCRIPT_DEBUG ? '' : '.min' ) . '.js', array( 'jquery' ), SNB_VERSION, TRUE );
        wp_localize_script( 
            'snbvars', 
            'snbvars', 
            array( 
                'ajaxurl'       => admin_url( 'admin-ajax.php' ),
                'minEvents'     => SNB_MIN_EVENTS > 2 ? SNB_MIN_EVENTS : 2,
                'disableSubmit' => SNB_DISABLE_SUBMIT,
                'verify'        => wp_create_nonce( 'snbvars' ),
            ) 
        );
    }
                   
    /**
     * 1. verify nonce 
     * 2. check time elapsed less than 1 hour ( FIXME: use WP nonce to combine nonce and timestamp? )
     * 3. create unique id; 
     * 4. create hash of event, timestamp and unique id and save with IP + unique id as key
     * 5. if all looks good, respond to browser with token.
     */
    function get_token(){
        if ( $this->is_spam_ip() )
            die();
        if ( check_ajax_referer( 'snbvars', 'verify' ) ):
            if ( isset( $_POST[ 'eventstr' ] ) ):
                $events = explode( ',', $_POST[ 'eventstr' ] );
                $time = array_pop( $events );
                $eventstr = implode( ',', $events ) . ',';
                // within the hour - js time in millisecs
                $elapsed = time() - ( $time / 1000 );
                if ( $elapsed < 3600 ):
                    $this->log_debug( __METHOD__, 'request: ' . print_r( $_REQUEST, TRUE ) );
                    //create unique id
                    $uniqid = uniqid();
                    // hash the events, timestamp and uniquid
                    $source = $eventstr . $time . $uniqid;
                    $hash   = hash( 'sha1', $source );
                    $this->log_debug( __METHOD__, 'hash: ' . $hash );
                    $this->log_debug( __METHOD__, 'UA: ' . $_SERVER[ 'HTTP_USER_AGENT' ] );
                    // save events, hash in transient uniquid
                    set_transient( $this->session_token( $uniqid ), $eventstr . ':' . $hash, 3600 ); // token valid for one hour
                    $this->log_debug( __METHOD__, 'token: ' . $uniqid );
                    // return the uniqid:hash
                    die( $uniqid );
                else:
                    $this->log_debug( __METHOD__, 'invalid token request (timestamp): ' . $elapsed . ' ' . print_r( $_REQUEST, TRUE ) );        
                endif;
            else:
                $this->log_debug( __METHOD__, 'invalid token request (no events): ' . print_r( $_REQUEST, TRUE ) );
            endif;
        else:
            $this->log_debug( __METHOD__, 'invalid token request (referrer): ' . $_SERVER[ 'HTTP_REFERER' ] . ' ' . print_r( $_REQUEST, TRUE ) );
        endif;
        $this->add_spam_ip();
        die();
    }
    
    /**
     * check remote ip against transient list of known offenders
     * return true if ip is in list, false otherwise.
     */
    function is_spam_ip(){
        if ( $spam_ips = get_transient( 'snb_spam_ips' ) ):
            if ( array_search( $_SERVER[ 'REMOTE_ADDR' ], $spam_ips ) ):
                $this->log_debug( __METHOD__, '***SPAM IP DETECTED!!!*** ' . $_SERVER[ 'REMOTE_ADDR' ] );
                return TRUE;
            endif;
        endif;
        return FALSE;
    }
    
    /**
     * add offending ip to transient list of ips. rotate oldest if greater than maximum number of ips
     */
    function add_spam_ip(){
        if ( !( $spam_ips = get_transient( 'snb_spam_ips' ) ) )
            $spam_ips = array();
        array_push( $spam_ips, $_SERVER[ 'REMOTE_ADDR' ] );
        $count = count( $spam_ips );
        if ( $count > SNB_MAX_SPAM_IPS )
            array_shift( $spam_ips );
        $this->log_debug( __METHOD__, 'adding spam ip ' . $count . ': ' . $_SERVER[ 'REMOTE_ADDR' ] );
        set_transient( 'snb_spam_ips', $spam_ips, SNB_SPAM_IP_LIFESPAN );
    }
    
    function reset_spam_ips(){
        if ( current_user_can( 'administrator' ) ):
            $this->log_debug( __METHOD__, 'clearing spam ips' );
            delete_transient( 'snb_spam_ips' );
        endif;
    }
    /**
     * link unique id to remote ip to prevent reuse of tokens across multiple requests
     */
    function session_token( $token ){
        return 'snbvars' . str_replace( '.', '', $_SERVER[ 'REMOTE_ADDR' ] ) . $token;
    }
    
    /**
     * spam tests are run after input validation,
     * so this won't affect legitimate users with invalid inputs.
     */
    function validate_session( $spam, $arg = 'contact-form-7' ) {
        if ( $this->is_spam_ip() )
            return TRUE;
        $this->log_debug( __METHOD__, 'referer: ' . $_SERVER[ 'HTTP_REFERER' ] );
        $this->log_debug( __METHOD__, 'UA: ' . $_SERVER[ 'HTTP_USER_AGENT' ] );
        if ( isset( $_POST[ 'snb-token' ] ) ):
            $this->log_debug( __METHOD__, 'snb-token: ' . $_POST[ 'snb-token' ] );
            list( $token, $theirhash, $start, $end ) = explode( ':', $_POST[ 'snb-token' ] );
            $this->log_debug( __METHOD__, 'token: ' . $token . ' theirhash: ' . $theirhash . ' start: ' . $start . ' end: ' . $end . ' time: ' . ( $end - $start ) );

            if ( $end - $start > 1500 ):
                if ( $session = get_transient( $this->session_token( $token ) ) ):
                    list( $events, $ourhash ) = explode( ':', $session );
                    if ( $theirhash === $ourhash ): // passed all token tests - now check events
                        $this->log_debug( __METHOD__, 'passed! events: ' . $events );
                        if ( preg_match( '{^f,.+?,[fm]}', $events ) ):
                            $this->log_debug( __METHOD__, ' invalid ' . $arg . ' event' );
                            //$this->log_debug( __METHOD__, 'full request: ' . print_r( $_REQUEST, TRUE ) );
                        else:
                            delete_transient( 'snbvars' . $token );
                            return $spam;
                        endif;
                    else:
                        $this->log_debug( __METHOD__, ' invalid ' . $arg . ' submission: mismatch ' . $theirhash . ' !== ' . $ourhash );
                    endif;
                else:
                    $this->log_debug( __METHOD__, ' invalid ' . $arg . ' submission: no session ' . $token );
                endif;
            else:
                $this->log_debug( __METHOD__, ' invalid ' . $arg . ' submission: suspiciously fast form completion: ' . ( $end - $start ) . 'ms' );
            endif;
        else:
            $this->log_debug( __METHOD__, ' invalid ' . $arg . ' submission: no token' );
        endif;
                
        $this->log_debug( __METHOD__, 'request: ' . print_r( $_REQUEST, TRUE ) );
        
        $this->add_spam_ip();
		return TRUE;
	}

    function log_debug( $fn, $msg ){
        if ( SNB_DEBUG ):
            @file_put_contents( 
                SNB_DIR . '/debug.txt', 
                implode( ':' , array(
                    current_time( 'mysql' ),
                    $_SERVER[ 'REMOTE_ADDR' ],
                    $_SERVER['REQUEST_URI'],
                    $fn,
                    $msg 
                ) ) . "\n", 
                FILE_APPEND );
        endif;
    }
}

defined( 'SNB_DEBUG' ) or define( 'SNB_DEBUG', FALSE );
defined( 'SNB_DISABLE_SUBMIT' ) || define( 'SNB_DISABLE_SUBMIT', FALSE );
defined( 'SNB_MIN_EVENTS' ) || define( 'SNB_MIN_EVENTS', 2 );
defined( 'SNB_SPAM_IP_LIFESPAN' ) or define( 'SNB_SPAM_IP_LIFESPAN', 60 * 60 * 24 * 30 ); // 30 days
defined( 'SNB_MAX_SPAM_IPS' ) or define( 'SNB_MAX_SPAM_IPS', 100 );
define( 'SNB_VERSION', '2.1.5' );
define( 'SNB_DIR', dirname( __FILE__ ) );

new SimpleNoBot();

endif;
